#include "vamp-hostsdk/PluginLoader.h"
#include "vamp-hostsdk/PluginInputDomainAdapter.h"
#include "vamp-hostsdk/PluginChannelAdapter.h"
#include "vamp-hostsdk/PluginBufferingAdapter.h"
#include "vamp-hostsdk/PluginHostAdapter.h"

#include "vamp.h"

#include "vamp-hostsdk/Files.h"

#include <fstream>

namespace Vamp {
	
namespace HostExt {

class PluginLoader::Impl
{
public:
    Impl();
    virtual ~Impl();

    PluginKeyList listPlugins();
    PluginKeyList listPluginsIn(std::vector<std::string>);
    PluginKeyList listPluginsNotIn(std::vector<std::string>);

    Plugin *loadPlugin(PluginKey key,
                       float inputSampleRate,
                       int adapterFlags);
    
    PluginKey composePluginKey(std::string libraryName, std::string identifier);

    PluginCategoryHierarchy getPluginCategory(PluginKey key);

    std::string getLibraryPathForPlugin(PluginKey key);

    static void setInstanceToClean(PluginLoader *instance);

protected:
    class PluginDeletionNotifyAdapter : public PluginWrapper {
    public:
        PluginDeletionNotifyAdapter(Plugin *plugin, Impl *loader);
        virtual ~PluginDeletionNotifyAdapter();
    protected:
        Impl *m_loader;
    };

    class InstanceCleaner {
    public:
        InstanceCleaner() : m_instance(0) { }
        ~InstanceCleaner() { delete m_instance; }
        void setInstance(PluginLoader *instance) { m_instance = instance; }
    protected:
        PluginLoader *m_instance;
    };

    virtual void pluginDeleted(PluginDeletionNotifyAdapter *adapter);

    std::map<PluginKey, std::string> m_pluginLibraryNameMap;
    bool m_allPluginsEnumerated;

    struct Enumeration {
        enum { All, SinglePlugin, InLibraries, NotInLibraries } type;
        PluginKey key;
        std::vector<std::string> libraryNames;
        Enumeration() : type(All) { }
    };
    std::vector<std::string> listLibraryFilesFor(Enumeration);

    /// Populate m_pluginLibraryNameMap and return a list of the keys
    /// that were added to it
    std::vector<PluginKey> enumeratePlugins(Enumeration);

    std::map<PluginKey, PluginCategoryHierarchy> m_taxonomy;
    void generateTaxonomy();

    std::map<Plugin *, void *> m_pluginLibraryHandleMap;

    bool decomposePluginKey(PluginKey key,
                            std::string &libraryName, std::string &identifier);

    static InstanceCleaner m_cleaner;
};

PluginLoader *
PluginLoader::m_instance = 0;

PluginLoader::Impl::InstanceCleaner
PluginLoader::Impl::m_cleaner;

PluginLoader::PluginLoader()
{
    m_impl = new Impl();
}

PluginLoader::~PluginLoader()
{
    delete m_impl;
}

PluginLoader *
PluginLoader::getInstance()
{
    if (!m_instance) {
        // The cleaner doesn't own the instance, because we leave the
        // instance pointer in the base class for binary backwards
        // compatibility reasons and to avoid waste
        m_instance = new PluginLoader();
        Impl::setInstanceToClean(m_instance);
    }
    return m_instance;
}

PluginLoader::PluginKeyList
PluginLoader::listPlugins() 
{
    return m_impl->listPlugins();
}

PluginLoader::PluginKeyList
PluginLoader::listPluginsIn(std::vector<std::string> libs) 
{
    return m_impl->listPluginsIn(libs);
}

PluginLoader::PluginKeyList
PluginLoader::listPluginsNotIn(std::vector<std::string> libs) 
{
    return m_impl->listPluginsNotIn(libs);
}

Plugin *
PluginLoader::loadPlugin(PluginKey key,
                         float inputSampleRate,
                         int adapterFlags)
{
    return m_impl->loadPlugin(key, inputSampleRate, adapterFlags);
}

PluginLoader::PluginKey
PluginLoader::composePluginKey(std::string libraryName, std::string identifier) 
{
    return m_impl->composePluginKey(libraryName, identifier);
}

PluginLoader::PluginCategoryHierarchy
PluginLoader::getPluginCategory(PluginKey key)
{
    return m_impl->getPluginCategory(key);
}

    std::string
PluginLoader::getLibraryPathForPlugin(PluginKey key)
{
    return m_impl->getLibraryPathForPlugin(key);
}
 
PluginLoader::Impl::Impl() :
    m_allPluginsEnumerated(false)
{
}

PluginLoader::Impl::~Impl()
{
}

void
PluginLoader::Impl::setInstanceToClean(PluginLoader *instance)
{
    m_cleaner.setInstance(instance);
}

PluginLoader::PluginKeyList
PluginLoader::Impl::listPlugins() 
{
    if (!m_allPluginsEnumerated) enumeratePlugins(Enumeration());

    std::vector<PluginKey> plugins;
    for (std::map<PluginKey, std::string>::const_iterator i =
             m_pluginLibraryNameMap.begin();
         i != m_pluginLibraryNameMap.end(); ++i) {
        plugins.push_back(i->first);
    }

    return plugins;
}

PluginLoader::PluginKeyList
PluginLoader::Impl::listPluginsIn(std::vector<std::string> libs) 
{
    Enumeration enumeration;
    enumeration.type = Enumeration::InLibraries;
    enumeration.libraryNames = libs;
    return enumeratePlugins(enumeration);
}

PluginLoader::PluginKeyList
PluginLoader::Impl::listPluginsNotIn(std::vector<std::string> libs) 
{
    Enumeration enumeration;
    enumeration.type = Enumeration::NotInLibraries;
    enumeration.libraryNames = libs;
    return enumeratePlugins(enumeration);
}

    std::vector<std::string>
PluginLoader::Impl::listLibraryFilesFor(Enumeration enumeration)
{
    Files::Filter filter;
    
    switch (enumeration.type) {

    case Enumeration::All:
        filter.type = Files::Filter::All;
        break;

    case Enumeration::SinglePlugin:
    {
        std::string libraryName, identifier;
        if (!decomposePluginKey(enumeration.key, libraryName, identifier)) {
            std::cerr << "WARNING: Vamp::HostExt::PluginLoader: "
                      << "Invalid plugin key \"" << enumeration.key
                      << "\" in enumerate" << std::endl;
            return std::vector<std::string>();
        }
        filter.type = Files::Filter::Matching;
        filter.libraryNames.clear();
        filter.libraryNames.push_back(libraryName);
        break;
    }

    case Enumeration::InLibraries:
        filter.type = Files::Filter::Matching;
        filter.libraryNames = enumeration.libraryNames;
        break;

    case Enumeration::NotInLibraries:
        filter.type = Files::Filter::NotMatching;
        filter.libraryNames = enumeration.libraryNames;
        break;
    }
        
    return Files::listLibraryFilesMatching(filter);
}

    std::vector<PluginLoader::PluginKey>
PluginLoader::Impl::enumeratePlugins(Enumeration enumeration)
{
    std::string libraryName, identifier;
    if (enumeration.type == Enumeration::SinglePlugin) {
        decomposePluginKey(enumeration.key, libraryName, identifier);
    }
    
    std::vector<std::string> fullPaths = listLibraryFilesFor(enumeration);

    // For these we should warn if a plugin can be loaded from a library
    bool specific = (enumeration.type == Enumeration::SinglePlugin ||
                     enumeration.type == Enumeration::InLibraries);

    std::vector<PluginKey> added;
    
    for (size_t i = 0; i < fullPaths.size(); ++i) {

        std::string fullPath = fullPaths[i];
        void *handle = Files::loadLibrary(fullPath);
        if (!handle) continue;
            
        VampGetPluginDescriptorFunction fn =
            (VampGetPluginDescriptorFunction)Files::lookupInLibrary
            (handle, "vampGetPluginDescriptor");
            
        if (!fn) {
            if (specific) {
                std::cerr << "Vamp::HostExt::PluginLoader: "
                    << "No vampGetPluginDescriptor function found in library \""
                     << fullPath << "\"" << std::endl;
            }
            Files::unloadLibrary(handle);
            continue;
        }
            
        int index = 0;
        const VampPluginDescriptor *descriptor = 0;
        bool found = false;
            
        while ((descriptor = fn(VAMP_API_VERSION, index))) {
            ++index;
            if (identifier != "") {
                if (descriptor->identifier != identifier) {
                    continue;
                }
            }
            found = true;
            PluginKey key = composePluginKey(fullPath, descriptor->identifier);
            if (m_pluginLibraryNameMap.find(key) ==
                m_pluginLibraryNameMap.end()) {
                m_pluginLibraryNameMap[key] = fullPath;
            }
            added.push_back(key);
        }

        if (!found && specific) {
            std::cerr << "Vamp::HostExt::PluginLoader: Plugin \""
                 << identifier << "\" not found in library \""
                 << fullPath << "\"" << std::endl;
        }
            
        Files::unloadLibrary(handle);
    }

    if (enumeration.type == Enumeration::All) {
        m_allPluginsEnumerated = true;
    }

    return added;
}

PluginLoader::PluginKey
PluginLoader::Impl::composePluginKey(std::string libraryName, std::string identifier)
{
    std::string basename = Files::lcBasename(libraryName);
    return basename + ":" + identifier;
}

bool
PluginLoader::Impl::decomposePluginKey(PluginKey key,
                                       std::string &libraryName,
                                       std::string &identifier)
{
    std::string::size_type ki = key.find(':');
    if (ki == std::string::npos) {
        return false;
    }

    libraryName = key.substr(0, ki);
    identifier = key.substr(ki + 1);
    return true;
}

PluginLoader::PluginCategoryHierarchy
PluginLoader::Impl::getPluginCategory(PluginKey plugin)
{
    if (m_taxonomy.empty()) generateTaxonomy();
    if (m_taxonomy.find(plugin) == m_taxonomy.end()) {
        return PluginCategoryHierarchy();
    }
    return m_taxonomy[plugin];
}

    std::string
PluginLoader::Impl::getLibraryPathForPlugin(PluginKey plugin)
{
    if (m_pluginLibraryNameMap.find(plugin) == m_pluginLibraryNameMap.end()) {
        if (m_allPluginsEnumerated) return "";
        Enumeration enumeration;
        enumeration.type = Enumeration::SinglePlugin;
        enumeration.key = plugin;
        enumeratePlugins(enumeration);
    }
    if (m_pluginLibraryNameMap.find(plugin) == m_pluginLibraryNameMap.end()) {
        return "";
    }
    return m_pluginLibraryNameMap[plugin];
}    

Plugin *
PluginLoader::Impl::loadPlugin(PluginKey key,
                               float inputSampleRate, int adapterFlags)
{
    std::string libname, identifier;
    if (!decomposePluginKey(key, libname, identifier)) {
        std::cerr << "Vamp::HostExt::PluginLoader: Invalid plugin key \""
                  << key << "\" in loadPlugin" << std::endl;
        return 0;
    }
        
    std::string fullPath = getLibraryPathForPlugin(key);
    if (fullPath == "") {
        std::cerr << "Vamp::HostExt::PluginLoader: No library found in Vamp path for plugin \"" << key << "\"" << std::endl;
        return 0;
    }
    
    void *handle = Files::loadLibrary(fullPath);
    if (!handle) return 0;
    
    VampGetPluginDescriptorFunction fn =
        (VampGetPluginDescriptorFunction)Files::lookupInLibrary
        (handle, "vampGetPluginDescriptor");

    if (!fn) {
        std::cerr << "Vamp::HostExt::PluginLoader: No vampGetPluginDescriptor function found in library \""
             << fullPath << "\"" << std::endl;
        Files::unloadLibrary(handle);
        return 0;
    }

    int index = 0;
    const VampPluginDescriptor *descriptor = 0;

    while ((descriptor = fn(VAMP_API_VERSION, index))) {

        if (std::string(descriptor->identifier) == identifier) {

            Vamp::PluginHostAdapter *plugin =
                new Vamp::PluginHostAdapter(descriptor, inputSampleRate);

            Plugin *adapter = new PluginDeletionNotifyAdapter(plugin, this);

            m_pluginLibraryHandleMap[adapter] = handle;

            if (adapterFlags & ADAPT_INPUT_DOMAIN) {
                if (adapter->getInputDomain() == Plugin::FrequencyDomain) {
                    adapter = new PluginInputDomainAdapter(adapter);
                }
            }

            if (adapterFlags & ADAPT_BUFFER_SIZE) {
                adapter = new PluginBufferingAdapter(adapter);
            }

            if (adapterFlags & ADAPT_CHANNEL_COUNT) {
                adapter = new PluginChannelAdapter(adapter);
            }

            return adapter;
        }

        ++index;
    }

    std::cerr << "Vamp::HostExt::PluginLoader: Plugin \""
         << identifier << "\" not found in library \""
         << fullPath << "\"" << std::endl;

    return 0;
}

void
PluginLoader::Impl::generateTaxonomy()
{
//    cerr << "PluginLoader::Impl::generateTaxonomy" << endl;

    std::vector<std::string> path = PluginHostAdapter::getPluginPath();
    std::string libfragment = "/lib/";
    std::vector<std::string> catpath;

    std::string suffix = "cat";

    for (std::vector<std::string>::iterator i = path.begin();
         i != path.end(); ++i) {

        // It doesn't matter that we're using literal forward-slash in
        // this bit, as it's only relevant if the path contains
        // "/lib/", which is only meaningful and only plausible on
        // systems with forward-slash delimiters
        
        std::string dir = *i;
        std::string::size_type li = dir.find(libfragment);

        if (li != std::string::npos) {
            catpath.push_back
                (dir.substr(0, li)
                 + "/share/"
                 + dir.substr(li + libfragment.length()));
        }

        catpath.push_back(dir);
    }

    char buffer[1024];

    for (std::vector<std::string>::iterator i = catpath.begin();
         i != catpath.end(); ++i) {
        
        std::vector<std::string> files = Files::listFiles(*i, suffix);

        for (std::vector<std::string>::iterator fi = files.begin();
             fi != files.end(); ++fi) {

            std::string filepath = Files::splicePath(*i, *fi);
            std::ifstream is(filepath.c_str(), std::ifstream::in | std::ifstream::binary);

            if (is.fail()) {
//                cerr << "failed to open: " << filepath << endl;
                continue;
            }

//            cerr << "opened: " << filepath << endl;

            while (!!is.getline(buffer, 1024)) {

                std::string line(buffer);

//                cerr << "line = " << line << endl;

                std::string::size_type di = line.find("::");
                if (di == std::string::npos) continue;

                std::string id = line.substr(0, di);
                std::string encodedCat = line.substr(di + 2);

                if (id.substr(0, 5) != "vamp:") continue;
                id = id.substr(5);

                while (encodedCat.length() >= 1 &&
                       encodedCat[encodedCat.length()-1] == '\r') {
                    encodedCat = encodedCat.substr(0, encodedCat.length()-1);
                }

//                cerr << "id = " << id << ", cat = " << encodedCat << endl;

                PluginCategoryHierarchy category;
                std::string::size_type ai;
                while ((ai = encodedCat.find(" > ")) != std::string::npos) {
                    category.push_back(encodedCat.substr(0, ai));
                    encodedCat = encodedCat.substr(ai + 3);
                }
                if (encodedCat != "") category.push_back(encodedCat);

                m_taxonomy[id] = category;
            }
        }
    }
}    

void
PluginLoader::Impl::pluginDeleted(PluginDeletionNotifyAdapter *adapter)
{
    void *handle = m_pluginLibraryHandleMap[adapter];
    if (!handle) return;

    m_pluginLibraryHandleMap.erase(adapter);

    for (auto h: m_pluginLibraryHandleMap) {
        if (h.second == handle) {
            // still in use
            return;
        }
    }
    
    Files::unloadLibrary(handle);
}

PluginLoader::Impl::PluginDeletionNotifyAdapter::PluginDeletionNotifyAdapter(Plugin *plugin,
                                                                             Impl *loader) :
    PluginWrapper(plugin),
    m_loader(loader)
{
}

PluginLoader::Impl::PluginDeletionNotifyAdapter::~PluginDeletionNotifyAdapter()
{
    // We need to delete the plugin before calling pluginDeleted, as
    // the delete call may require calling through to the descriptor
    // (for e.g. cleanup) but pluginDeleted may unload the required
    // library for the call.  To prevent a double deletion when our
    // parent's destructor runs (after this one), be sure to set
    // m_plugin to 0 after deletion.
    delete m_plugin;
    m_plugin = 0;

    if (m_loader) m_loader->pluginDeleted(this);
}

}

}
