#pragma once

#include "vamp-sdk/Plugin.h"

namespace Vamp::HostExt {

/**
 * \class PluginWrapper PluginWrapper.h <vamp-hostsdk/PluginWrapper.h>
 * 
 * PluginWrapper is a simple base class for adapter plugins.  It takes
 * a pointer to a "to be wrapped" Vamp plugin on construction, and
 * provides implementations of all the Vamp plugin methods that simply
 * delegate through to the wrapped plugin.  A subclass can therefore
 * override only the methods that are meaningful for the particular
 * adapter.
 *
 * \note This class was introduced in version 1.1 of the Vamp plugin SDK.
 */

class PluginWrapper : public Plugin
{
public:
    virtual ~PluginWrapper();
    
    bool initialise(size_t channels, size_t stepSize, size_t blockSize);
    void reset();

    InputDomain getInputDomain() const;

    unsigned int getVampApiVersion() const;
    std::string getIdentifier() const;
    std::string getName() const;
    std::string getDescription() const;
    std::string getMaker() const;
    int getPluginVersion() const;
    std::string getCopyright() const;

    ParameterList getParameterDescriptors() const;
    float getParameter(std::string) const;
    void setParameter(std::string, float);

    ProgramList getPrograms() const;
    std::string getCurrentProgram() const;
    void selectProgram(std::string);

    size_t getPreferredStepSize() const;
    size_t getPreferredBlockSize() const;

    size_t getMinChannelCount() const;
    size_t getMaxChannelCount() const;

    OutputList getOutputDescriptors() const;

    FeatureSet process(const float *const *inputBuffers, RealTime timestamp);

    FeatureSet getRemainingFeatures();

    /**
     * Return a pointer to the plugin wrapper of type WrapperType
     * surrounding this wrapper's plugin, if present.
     *
     * This is useful in situations where a plugin is wrapped by
     * multiple different wrappers (one inside another) and the host
     * wants to call some wrapper-specific function on one of the
     * layers without having to care about the order in which they are
     * wrapped.  For example, the plugin returned by
     * PluginLoader::loadPlugin may have more than one wrapper; if the
     * host wanted to query or fine-tune some property of one of them,
     * it would be hard to do so without knowing the order of the
     * wrappers.  This function therefore gives direct access to the
     * wrapper of a particular type.
     */
    template <typename WrapperType>
    WrapperType *getWrapper() {
        WrapperType *w = dynamic_cast<WrapperType *>(this);
        if (w) return w;
        PluginWrapper *pw = dynamic_cast<PluginWrapper *>(m_plugin);
        if (pw) return pw->getWrapper<WrapperType>();
        return 0;
    }

protected:
    PluginWrapper(Plugin *plugin); // I take ownership of plugin
    Plugin *m_plugin;
};

}

