#pragma once

#include <map>
#include "vamp.h"

#include "vamp-sdk/Plugin.h"

namespace Vamp {

/**
 * \class PluginAdapterBase PluginAdapter.h <vamp-sdk/PluginAdapter.h>
 * 
 * PluginAdapter and PluginAdapterBase provide a wrapper class that a
 * plugin library can use to make its C++ Vamp::Plugin objects
 * available through the Vamp C API.
 * 
 * Almost all Vamp plugin libraries will want to make use of this.  To
 * do so, all they need to do is declare a PluginAdapter<T> for each
 * plugin class T in their library.  It's very simple, and you need to
 * know absolutely nothing about how it works in order to use it.
 * Just cut and paste from an existing plugin's discovery function.
 * \see vampGetPluginDescriptor
 */

class PluginAdapterBase
{
public:
    virtual ~PluginAdapterBase();

    /**
     * Return a VampPluginDescriptor describing the plugin that is
     * wrapped by this adapter.
     */
    const VampPluginDescriptor *getDescriptor();

protected:
    PluginAdapterBase();

    virtual Plugin *createPlugin(float inputSampleRate) = 0;

    class Impl;
    Impl *m_impl;
};

/**
 * \class PluginAdapter PluginAdapter.h <vamp-sdk/PluginAdapter.h>
 * 
 * PluginAdapter turns a PluginAdapterBase into a specific wrapper for
 * a particular plugin implementation.
 *
 * See PluginAdapterBase.
 */

template <typename P>
class PluginAdapter : public PluginAdapterBase
{
public:
    PluginAdapter() : PluginAdapterBase() { }
    virtual ~PluginAdapter() { }

protected:
    Plugin *createPlugin(float inputSampleRate) {
        P *p = new P(inputSampleRate);
        Plugin *plugin = dynamic_cast<Plugin *>(p);
        if (!plugin) {
            std::cerr << "ERROR: PluginAdapter::createPlugin: "
                      << "Template type is not a plugin!"
                      << std::endl;
            delete p;
            return 0;
        }
        return plugin;
    }
};
    
}
